// Licensed to Elasticsearch B.V. under one or more contributor
// license agreements. See the NOTICE file distributed with
// this work for additional information regarding copyright
// ownership. Elasticsearch B.V. licenses this file to you under
// the Apache License, Version 2.0 (the "License"); you may
// not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

package intake

import (
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

func TestCacheInitFails(t *testing.T) {
	for _, test := range []struct {
		size  int
		limit int
	}{
		{-1, 1},
		{0, 1},
		{1, -1},
	} {
		c, err := NewRateLimitLRUCache(test.size, test.limit, 3)
		assert.Error(t, err)
		assert.Nil(t, c)
	}
}

func TestCacheEviction(t *testing.T) {
	cacheSize := 2
	limit := 1 //multiplied times BurstMultiplier 3

	rlc, err := NewRateLimitLRUCache(cacheSize, limit, 3)
	require.NoError(t, err)

	// add new limiter
	rlA, _ := rlc.RateLimiter("a")
	rlA.AllowN(time.Now(), 3)

	// add new limiter
	rlB, _ := rlc.RateLimiter("b")
	rlB.AllowN(time.Now(), 2)

	// reuse evicted limiter rlA
	rlC, _ := rlc.RateLimiter("c")
	assert.False(t, rlC.Allow())
	assert.Equal(t, rlC, rlc.evictedLimiter)

	// reuse evicted limiter rlB
	rlD, _ := rlc.RateLimiter("a")
	assert.True(t, rlD.Allow())
	assert.False(t, rlD.Allow())
	assert.Equal(t, rlD, rlc.evictedLimiter)
	// check that limiter are independent
	assert.True(t, rlD != rlC)
	rlc.evictedLimiter = nil
	assert.NotNil(t, rlD)
	assert.NotNil(t, rlC)
}

func TestCacheOk(t *testing.T) {
	var rlc *RateLimitLRUCache
	_, ok := rlc.RateLimiter("a")
	assert.False(t, ok)

	var cache = func() *RateLimitLRUCache {
		rlc, err := NewRateLimitLRUCache(1, 1, 1)
		require.NoError(t, err)
		return rlc
	}

	rlc = cache()
	rlc.limit = -1
	_, ok = rlc.RateLimiter("a")
	assert.False(t, ok)

	rlc = cache()
	rlc.cache = nil
	_, ok = rlc.RateLimiter("a")
	assert.False(t, ok)

	rlc = cache()
	_, ok = rlc.RateLimiter("a")
	assert.True(t, ok)
}
